@@ -1,6 +1,7 @@
 """The default JSONPath parser."""
 from __future__ import annotations
 
+import codecs
 import re
 from typing import TYPE_CHECKING
 from typing import Callable
@@ -11,6 +12,7 @@ from typing import Optional
 from typing import Union
 
 from .exceptions import JSONPathSyntaxError
+from .filter import CURRENT_KEY
 from .filter import FALSE
 from .filter import NIL
 from .filter import TRUE
@@ -19,6 +21,7 @@ from .filter import BooleanExpression
 from .filter import FilterContextPath
 from .filter import FilterExpression
 from .filter import FloatLiteral
+from .filter import FunctionExtension
 from .filter import InfixExpression
 from .filter import IntegerLiteral
 from .filter import ListLiteral
@@ -31,6 +34,7 @@ from .path import JSONPath
 from .selectors import Filter
 from .selectors import IndexSelector
 from .selectors import JSONPathSelector
+from .selectors import KeysSelector
 from .selectors import ListSelector
 from .selectors import PropertySelector
 from .selectors import RecursiveDescentSelector
@@ -48,15 +52,17 @@ from .token import TOKEN_FILTER_CONTEXT
 from .token import TOKEN_FILTER_END
 from .token import TOKEN_FILTER_START
 from .token import TOKEN_FLOAT
+from .token import TOKEN_FUNCTION
 from .token import TOKEN_GE
 from .token import TOKEN_GT
 from .token import TOKEN_IN
 from .token import TOKEN_INDEX
 from .token import TOKEN_INT
 from .token import TOKEN_INTERSECTION
+from .token import TOKEN_KEY
+from .token import TOKEN_KEYS
 from .token import TOKEN_LE
 from .token import TOKEN_LG
-from .token import TOKEN_LIST_END
 from .token import TOKEN_LIST_START
 from .token import TOKEN_LPAREN
 from .token import TOKEN_LT
@@ -68,6 +74,7 @@ from .token import TOKEN_NOT
 from .token import TOKEN_NULL
 from .token import TOKEN_OR
 from .token import TOKEN_PROPERTY
+from .token import TOKEN_RBRACKET
 from .token import TOKEN_RE
 from .token import TOKEN_RE_FLAGS
 from .token import TOKEN_RE_PATTERN
@@ -87,9 +94,46 @@ if TYPE_CHECKING:
     from .env import JSONPathEnvironment
     from .stream import TokenStream
 
+# ruff: noqa: D102
+
+INVALID_NAME_SELECTOR_CHARS = [
+    "\x00",
+    "\x01",
+    "\x02",
+    "\x03",
+    "\x04",
+    "\x05",
+    "\x06",
+    "\x07",
+    "\x08",
+    "\t",
+    "\n",
+    "\x0b",
+    "\x0c",
+    "\r",
+    "\x0e",
+    "\x0f",
+    "\x10",
+    "\x11",
+    "\x12",
+    "\x13",
+    "\x14",
+    "\x15",
+    "\x16",
+    "\x17",
+    "\x18",
+    "\x19",
+    "\x1a",
+    "\x1b",
+    "\x1c",
+    "\x1d",
+    "\x1e",
+    "\x1f",
+]
+
 
 class Parser:
-    """A JSONPath expression parser bound to a JSONPathEnvironment."""
+    """A JSONPath parser bound to a JSONPathEnvironment."""
 
     PRECEDENCE_LOWEST = 1
     PRECEDENCE_LOGICALRIGHT = 3
@@ -144,6 +188,11 @@ class Parser:
         "s": re.S,
     }
 
+    _INVALID_NAME_SELECTOR_CHARS = f"[{''.join(INVALID_NAME_SELECTOR_CHARS)}]"
+    RE_INVALID_NAME_SELECTOR = re.compile(
+        rf'(?:(?!(?<!\\)"){_INVALID_NAME_SELECTOR_CHARS})'
+    )
+
     def __init__(self, *, env: JSONPathEnvironment) -> None:
         self.env = env
 
@@ -151,6 +200,7 @@ class Parser:
             TOKEN_FALSE: self.parse_boolean,
             TOKEN_FLOAT: self.parse_float_literal,
             TOKEN_INT: self.parse_integer_literal,
+            TOKEN_KEY: self.parse_current_key,
             TOKEN_LIST_START: self.parse_list_literal,
             TOKEN_LPAREN: self.parse_grouped_expression,
             TOKEN_MISSING: self.parse_undefined,
@@ -165,6 +215,7 @@ class Parser:
             TOKEN_STRING: self.parse_string_literal,
             TOKEN_TRUE: self.parse_boolean,
             TOKEN_UNDEFINED: self.parse_undefined,
+            TOKEN_FUNCTION: self.parse_function_extension,
         }
 
         self.list_item_map: Dict[str, Callable[[TokenStream], FilterExpression]] = {
@@ -178,6 +229,22 @@ class Parser:
             TOKEN_TRUE: self.parse_boolean,
         }
 
+        self.function_argument_map: Dict[
+            str, Callable[[TokenStream], FilterExpression]
+        ] = {
+            TOKEN_FALSE: self.parse_boolean,
+            TOKEN_FLOAT: self.parse_float_literal,
+            TOKEN_INT: self.parse_integer_literal,
+            TOKEN_KEY: self.parse_current_key,
+            TOKEN_NIL: self.parse_nil,
+            TOKEN_NONE: self.parse_nil,
+            TOKEN_NULL: self.parse_nil,
+            TOKEN_STRING: self.parse_string_literal,
+            TOKEN_TRUE: self.parse_boolean,
+            TOKEN_ROOT: self.parse_root_path,
+            TOKEN_SELF: self.parse_self_path,
+        }
+
     def parse(self, stream: TokenStream) -> Iterable[JSONPathSelector]:
         """Parse a JSONPath from a stream of tokens."""
         if stream.current.kind == TOKEN_ROOT:
@@ -205,11 +272,7 @@ class Parser:
                     name=stream.current.value,
                 )
             elif stream.current.kind == TOKEN_INDEX:
-                yield IndexSelector(
-                    env=self.env,
-                    token=stream.current,
-                    index=int(stream.current.value),
-                )
+                yield self.parse_index(stream)
             elif stream.current.kind == TOKEN_SLICE_START:
                 yield self.parse_slice(stream)
             elif stream.current.kind == TOKEN_WILD:
@@ -217,6 +280,11 @@ class Parser:
                     env=self.env,
                     token=stream.current,
                 )
+            elif stream.current.kind == TOKEN_KEYS:
+                yield KeysSelector(
+                    env=self.env,
+                    token=stream.current,
+                )
             elif stream.current.kind == TOKEN_DDOT:
                 yield RecursiveDescentSelector(
                     env=self.env,
@@ -224,7 +292,7 @@ class Parser:
                 )
             elif stream.current.kind == TOKEN_LIST_START:
                 yield self.parse_selector_list(stream)
-            elif stream.current.kind == TOKEN_FILTER_START and not in_filter:
+            elif stream.current.kind == TOKEN_FILTER_START:
                 yield self.parse_filter(stream)
             else:
                 if in_filter:
@@ -233,6 +301,20 @@ class Parser:
 
             stream.next_token()
 
+    def parse_index(self, stream: TokenStream) -> IndexSelector:
+        """Parse an index selector from a stream of tokens."""
+        if (
+            len(stream.current.value) > 1 and stream.current.value.startswith("0")
+        ) or stream.current.value.startswith("-0"):
+            raise JSONPathSyntaxError(
+                "leading zero in index selector", token=stream.current
+            )
+        return IndexSelector(
+            env=self.env,
+            token=stream.current,
+            index=int(stream.current.value),
+        )
+
     def parse_slice(self, stream: TokenStream) -> SliceSelector:
         """Parse a slice JSONPath expression from a stream of tokens."""
         start_token = stream.next_token()
@@ -265,12 +347,19 @@ class Parser:
         )
 
     def parse_selector_list(self, stream: TokenStream) -> ListSelector:
-        """Parse a comma separated list JSONPath selectors from a stream
-        of tokens."""
+        """Parse a comma separated list JSONPath selectors from a stream of tokens."""
         tok = stream.next_token()
-        list_items: List[Union[IndexSelector, PropertySelector, SliceSelector]] = []
+        list_items: List[
+            Union[
+                IndexSelector,
+                KeysSelector,
+                PropertySelector,
+                SliceSelector,
+                WildSelector,
+            ]
+        ] = []
 
-        while stream.current.kind != TOKEN_LIST_END:
+        while stream.current.kind != TOKEN_RBRACKET:
             if stream.current.kind == TOKEN_INT:
                 list_items.append(
                     IndexSelector(
@@ -279,28 +368,63 @@ class Parser:
                         index=int(stream.current.value),
                     )
                 )
-            elif stream.current.kind in (TOKEN_BARE_PROPERTY, TOKEN_STRING):
+            elif stream.current.kind == TOKEN_BARE_PROPERTY:
                 list_items.append(
                     PropertySelector(
                         env=self.env,
                         token=stream.current,
                         name=stream.current.value,
+                    ),
+                )
+            elif stream.current.kind == TOKEN_KEYS:
+                list_items.append(
+                    KeysSelector(
+                        env=self.env,
+                        token=stream.current,
                     )
                 )
+            elif stream.current.kind == TOKEN_STRING:
+                if self.RE_INVALID_NAME_SELECTOR.search(stream.current.value):
+                    raise JSONPathSyntaxError(
+                        f"invalid name selector {stream.current.value!r}",
+                        token=stream.current,
+                    )
+
+                name = (
+                    codecs.decode(
+                        stream.current.value.replace("\\/", "/"), "unicode-escape"
+                    )
+                    .encode("utf-16", "surrogatepass")
+                    .decode("utf-16")
+                )
+
+                list_items.append(
+                    PropertySelector(
+                        env=self.env,
+                        token=stream.current,
+                        name=name,
+                    ),
+                )
             elif stream.current.kind == TOKEN_SLICE_START:
                 list_items.append(self.parse_slice(stream))
-            elif stream.current.kind == TOKEN_EOF:
+            elif stream.current.kind == TOKEN_WILD:
+                list_items.append(WildSelector(env=self.env, token=stream.current))
+
+            if stream.peek.kind == TOKEN_EOF:
                 raise JSONPathSyntaxError(
                     "unexpected end of selector list",
                     token=stream.current,
                 )
 
-            if stream.peek.kind != TOKEN_LIST_END:
+            if stream.peek.kind != TOKEN_RBRACKET:
+                stream.expect_peek(TOKEN_COMMA)
                 stream.next_token()
 
             stream.next_token()
 
-        # TODO: is an empty list an error?
+        if not list_items:
+            raise JSONPathSyntaxError("empty segment", token=tok)
+
         return ListSelector(env=self.env, token=tok, items=list_items)
 
     def parse_filter(self, stream: TokenStream) -> Filter:
@@ -311,7 +435,7 @@ class Parser:
             raise JSONPathSyntaxError("unbalanced ')'", token=stream.current)
 
         stream.next_token()
-        stream.expect(TOKEN_FILTER_END)
+        stream.expect(TOKEN_FILTER_END, TOKEN_RBRACKET)
         return Filter(env=self.env, token=tok, expression=expr)
 
     def parse_boolean(self, stream: TokenStream) -> FilterExpression:
@@ -361,13 +485,18 @@ class Parser:
         stream.next_token()
 
         while stream.current.kind != TOKEN_RPAREN:
-            if stream.current.kind in (TOKEN_EOF, TOKEN_FILTER_END):
+            if stream.current.kind == TOKEN_EOF:
                 raise JSONPathSyntaxError(
                     "unbalanced parentheses", token=stream.current
                 )
+            if stream.current.kind == TOKEN_FILTER_END:
+                # In some cases, an RPAREN followed by an RBRACKET can
+                # look like the long form "end of filter" token.
+                stream.push(stream.current)
+                break
             expr = self.parse_infix_expression(stream, expr)
 
-        stream.expect(TOKEN_RPAREN)
+        stream.expect(TOKEN_RPAREN, TOKEN_FILTER_END)
         return expr
 
     def parse_root_path(self, stream: TokenStream) -> FilterExpression:
@@ -382,6 +511,9 @@ class Parser:
             JSONPath(env=self.env, selectors=self.parse_path(stream, in_filter=True))
         )
 
+    def parse_current_key(self, _: TokenStream) -> FilterExpression:
+        return CURRENT_KEY
+
     def parse_filter_context_path(self, stream: TokenStream) -> FilterExpression:
         stream.next_token()
         return FilterContextPath(
@@ -401,7 +533,7 @@ class Parser:
         stream.next_token()
         list_items: List[FilterExpression] = []
 
-        while stream.current.kind != TOKEN_LIST_END:
+        while stream.current.kind != TOKEN_RBRACKET:
             try:
                 list_items.append(self.list_item_map[stream.current.kind](stream))
             except KeyError as err:
@@ -410,7 +542,7 @@ class Parser:
                     token=stream.current,
                 ) from err
 
-            if stream.peek.kind != TOKEN_LIST_END:
+            if stream.peek.kind != TOKEN_RBRACKET:
                 stream.expect_peek(TOKEN_COMMA)
                 stream.next_token()
 
@@ -418,13 +550,41 @@ class Parser:
 
         return ListLiteral(list_items)
 
+    def parse_function_extension(self, stream: TokenStream) -> FilterExpression:
+        function_arguments: List[FilterExpression] = []
+        tok = stream.next_token()
+
+        while stream.current.kind != TOKEN_RPAREN:
+            try:
+                function_arguments.append(
+                    self.function_argument_map[stream.current.kind](stream)
+                )
+            except KeyError as err:
+                raise JSONPathSyntaxError(
+                    f"unexpected {stream.current.value!r}",
+                    token=stream.current,
+                ) from err
+
+            if stream.peek.kind != TOKEN_RPAREN:
+                if stream.peek.kind == TOKEN_FILTER_END:
+                    break
+                stream.expect_peek(TOKEN_COMMA)
+                stream.next_token()
+
+            stream.next_token()
+
+        function_arguments = self.env.validate_function_extension_signature(
+            tok, function_arguments
+        )
+        return FunctionExtension(tok.value, function_arguments)
+
     def parse_filter_selector(
         self, stream: TokenStream, precedence: int = PRECEDENCE_LOWEST
     ) -> FilterExpression:
         try:
             left = self.token_map[stream.current.kind](stream)
         except KeyError as err:
-            if stream.current.kind in (TOKEN_EOF, TOKEN_FILTER_END):
+            if stream.current.kind in (TOKEN_EOF, TOKEN_FILTER_END, TOKEN_RBRACKET):
                 msg = "end of expression"
             else:
                 msg = repr(stream.current.value)
@@ -435,7 +595,7 @@ class Parser:
         while True:
             peek_kind = stream.peek.kind
             if (
-                peek_kind in (TOKEN_EOF, TOKEN_FILTER_END)
+                peek_kind in (TOKEN_EOF, TOKEN_FILTER_END, TOKEN_RBRACKET)
                 or self.PRECEDENCES.get(peek_kind, self.PRECEDENCE_LOWEST) < precedence
             ):
                 break
